#!/usr/bin/perl -I/usr/sausalito/perl
# $Id: generate_kvm_list.pl, v2.0.0-1 Sun 20 Nov 2011 08:15:56 PM CET mstauber Exp $
# Copyright 2006-2011 Stauber Multimedia Design. All rights reserved.

# This constructor / handler uses 'virsh' and 'vzlist' to poll just very basic information
# off the master node to get a general idea of which KVM VPS's are running.
#
# This information needs to be supplemented later on with extra data.
#

$whatami = "constructor";

use CCE;
$cce = new CCE;

use Sys::Virt;
use Sys::Virt::Domain;
#use Time::HiRes qw(time);
use XML::Simple;
use Data::Dumper;

if ($whatami eq "handler") {
    $cce->connectfd();
}
else {
    $cce->connectuds();
}

# Debugging switch:
$DEBUG = "0";

# Location of libvirt config files (directory - with trailing slash):
$cfgdir = '/etc/libvirt/qemu/';

# We are starting sane:
$bad_info_detected = '0';

# Parse existing VPS and shove their data into CCE:
&main;

$cce->bye('SUCCESS');
exit(0);

##
# Subs:
##

sub main {

    # Check which KVM's are pesent (if any):
    $hv = Sys::Virt->new(address => "", readonly => 1);

    $interval = @ARGV ? shift @ARGV : 1;
    $iterations = @ARGV ? shift @ARGV : 1;

    %states = (
	&Sys::Virt::Domain::STATE_NOSTATE => "nostate",
	&Sys::Virt::Domain::STATE_RUNNING => "running",
	&Sys::Virt::Domain::STATE_BLOCKED => "blocked",
	&Sys::Virt::Domain::STATE_PAUSED => "paused",
	&Sys::Virt::Domain::STATE_SHUTDOWN => "shutdown",
	&Sys::Virt::Domain::STATE_SHUTOFF => "shutoff",
	&Sys::Virt::Domain::STATE_CRASHED => "crashed",
	&Sys::Virt::Domain::STATE_RUNNING => "running",
    );

    %cpuTime;
    $sample;

    for ($i = 0 ; $i < $iterations ; $i++) {
	sleep $interval if $i;
	$now = time;   

    	@domains_running = $hv->list_domains();
	@domains_stopped = $hv->list_defined_domains();
        @domains = (@domains_running, @domains_stopped);

	@stats;

	foreach $domain (sort { $a->get_id <=> $b->get_id } @domains) {
    	    $uuid = $domain->get_uuid_string;

    	    $info = $domain->get_info;

	    $cpudelta = exists $cpuTime{$uuid} ? $info->{cpuTime} - $cpuTime{$uuid} : 0;
	    $timedelta = defined $sample ? ($now - $sample)*1000*1000*1000 :0;

	    $cpuTime{$uuid} = $info->{cpuTime};

	    $util = $timedelta > 0 ? $cpudelta * 100 / $timedelta : 0;

	    if ($domain->get_autostart() == "-1") {
		$autostart = "0";
	    }
	    else {
		$autostart = "1";
	    }
	    if ($domain->get_os_type() == "hvm") {
		$hvm = "1";
	    }
	    else {
		$hvm = "0";
	    }

    	    # Build Matrix:
    	    $vpsid = $domain->get_name;
    	    if ($VPSID[$vpsid][0] ne "-1") {
    		$VPSID[$vpsid][0] = $domain->get_id;
    	    }
    	    else {
    		$VPSID[$vpsid][0] = "";
    	    }
    	    $VPSID[$vpsid][1] = $domain->get_name;
    	    $VPSID[$vpsid][25] = $states{$info->{state}};
    	    $VPSID[$vpsid][28] = (sprintf "%d%%",$util);
    	    $VPSID[$vpsid][3] = $info->{memory};
    	    $VPSID[$vpsid][2] = $info->{nrVirtCpu};
    	    $VPSID[$vpsid][29] = $info->{maxMem};
    	    $VPSID[$vpsid][30] = $autostart;
    	    $VPSID[$vpsid][14] = $hvm;

	    # Get the rest from XML:
	    $vpsxml = $domain->get_xml_description();
	    $xml = new XML::Simple;
	    $data = $xml->XMLin("$vpsxml");

    	    $VPSID[$vpsid][4] = $data->{'os'}->{'type'}->{'arch'};
    	    $VPSID[$vpsid][8] = $data->{'os'}->{'type'}->{'machine'};
    	    $VPSID[$vpsid][31] = $data->{'devices'}->{'disk'}->[0]->{'source'}->{'file'};
    	    $VPSID[$vpsid][10] = $data->{'devices'}->{'disk'}->[0]->{'target'}->{'bus'};
    	    $VPSID[$vpsid][9] = $data->{'devices'}->{'disk'}->[0]->{'device'};
    	    $VPSID[$vpsid][19] = $data->{'devices'}->{'graphics'}->{'type'};
#    	    $VPSID[$vpsid][21] = $data->{'devices'}->{'graphics'}->{'listen'};
    	    $VPSID[$vpsid][21] = $data->{'devices'}->{'graphics'}->{'listen'}[0];
    	    $VPSID[$vpsid][22] = $data->{'devices'}->{'graphics'}->{'port'};
	    $disk = $domain->get_block_info($VPSID[$vpsid][31]);
	    $VPSID[$vpsid][11] = $disk->{'capacity'};
	    $VPSID[$vpsid][32] = $disk->{'allocation'};
    	    $VPSID[$vpsid][12] = $data->{'devices'}->{'interface'}->{'type'};
    	    $VPSID[$vpsid][15] = $data->{'type'};
    	    $VPSID[$vpsid][18] = $VPSID[$vpsid][30]; # hmpf ... have this twice.

	    if ($DEBUG == "1") {
		$xml->XMLout($data,
        	    KeepRoot   => 1,
        	    OutputFile => "/tmp/debug-$vpsid.xml",
        	    RootName => 'domain',
        	    XMLDecl    => "",
    		);
	    }

	    # VNC passwd cannot be obtained through $domain->get_xml_description();
	    # For that we parse the XML configs in '/etc/libvirt/qemu/', which we'd probably do right from the start instead.

	    # create object
	    $vncxml = new XML::Simple(ForceArray => 1, KeepRoot => 1, NoSort => 0);

	    # read XML file
	    $vncdata = $vncxml->XMLin("$cfgdir$vpsid.xml");
	    $VPSID[$vpsid][20] = $vncdata->{domain}->[0]->{'devices'}->[0]->{'graphics'}->[0]->{'passwd'};

	    # Check if the file containing the KVM disk is present:
	    if (-f $VPSID[$vpsid][31]) {
		$is_dir_present = "1";
		$VPSID[$vpsid][26] = "1";
	    }
	    else {
		$is_dir_present = "0";
		$VPSID[$vpsid][26] = "0";
	    }
	    # Check if the XML config file containing the KVM config is present:
	    $the_file = $cfgdir . $VPSID[$vpsid][1] . ".xml";
	    if (-f $the_file) {
		$VPSID[$vpsid][27] = "1";
	    }
	    else {
		$VPSID[$vpsid][27] = "0";
	    }

	    if ($DEBUG == "1") {
		print "vpsid " . $VPSID[$vpsid][0] . "\n";
		print "vpsname " . $VPSID[$vpsid][1] . "\n";
		print "status " . $VPSID[$vpsid][25] . "\n";
		print "cpuTime " . $VPSID[$vpsid][28] . "\n";
		print "memory " . $VPSID[$vpsid][3] . "\n";
		print "vcpus " . $VPSID[$vpsid][2] . "\n";
		print "autostart " . $VPSID[$vpsid][30] . "\n";
		print "hvm " . $VPSID[$vpsid][14] . "\n";
		print "arch " . $VPSID[$vpsid][4] . "\n";
		print "os_variant " . $VPSID[$vpsid][8] . "\n";
		print "disk " . $VPSID[$vpsid][31] . "\n";
		print "disk_device " . $VPSID[$vpsid][9] . "\n";
		print "disk_bus " . $VPSID[$vpsid][10] . "\n";
		print "disk_size " . $VPSID[$vpsid][11] . "\n";
		print "disk_alloc " . $VPSID[$vpsid][32] . "\n";
		print "graphics " . $VPSID[$vpsid][19] . "\n";
		print "vnc_listen " . $VPSID[$vpsid][21] . "\n";
		print "vnc_port " . $VPSID[$vpsid][22] . "\n";
		print "vnc_pass " . $VPSID[$vpsid][20] . "\n";
		print "is_dir_present " . $VPSID[$vpsid][26] . "\n";
		print "is_vps_present " . $VPSID[$vpsid][27] . "\n";
		print "network " . $VPSID[$vpsid][12] . "\n";
		print "virt_type " . $VPSID[$vpsid][15] . "\n";
		print "\n";
#		print Dumper($data);
	    }
	    # Write the data off into CODB:
	    &feedthemonster;
	}
	$sample = $now;
    }
}


# Subroutine that feeds the data into CCE:
sub feedthemonster {

	# This is where we actually populate CCE with the gathered info:
	$vpsname = $VPSID[$vpsid][1];
	$vpsid = $VPSID[$vpsid][1];

	# Check if the VPS is already present in CCE:
        @oids = $cce->find('KVM', {'vpsname' => $vpsname});

        if ($#oids == -1) {
                # Object not yet in CCE. Creating new one:
                #'vpsid' => $VPSID[$vpsid][0],
                ($ok) = $cce->create('KVM', {
                'vpsname' => $VPSID[$vpsid][1],
                'vcpus' => $VPSID[$vpsid][2],
                'memory' => $VPSID[$vpsid][3],
                'arch' => $VPSID[$vpsid][4],
                'os_type' => $VPSID[$vpsid][7],
                'os_variant' => $VPSID[$vpsid][8],
                'disk_device' => $VPSID[$vpsid][9],
                'disk_bus' => $VPSID[$vpsid][10],
                'disk_size' => $VPSID[$vpsid][11],
                'network' => $VPSID[$vpsid][12],
                'hvm' => $VPSID[$vpsid][14],
                'virt_type' => $VPSID[$vpsid][15],
                'autostart' => $VPSID[$vpsid][18],
                'graphics' => $VPSID[$vpsid][19],
                'vnc_listen' => $VPSID[$vpsid][21],
                'vnc_port' => $VPSID[$vpsid][22],
                'status' => $VPSID[$vpsid][25],
                'is_dir_present' => $VPSID[$vpsid][26],
                'is_vps_present' => $VPSID[$vpsid][27],
                'cpuTime' => $VPSID[$vpsid][28],
                'currMem' => $VPSID[$vpsid][29],
                'disk' => $VPSID[$vpsid][31],
                'disk_alloc' => $VPSID[$vpsid][32],
                'timestamp' => time()
                });
        }
        else {
	        # Object already present in CCE. Updating it.
                ($ok) = $cce->set($oids[0], '',{ 	# beware: CLASS vs. Namespace (see schema)! Our NameSpace is empty, hence: ''
                'vpsname' => $VPSID[$vpsid][1],
                'vcpus' => $VPSID[$vpsid][2],
                'memory' => $VPSID[$vpsid][3],
                'arch' => $VPSID[$vpsid][4],
                'os_type' => $VPSID[$vpsid][7],
                'os_variant' => $VPSID[$vpsid][8],
                'disk_device' => $VPSID[$vpsid][9],
                'disk_bus' => $VPSID[$vpsid][10],
                'disk_size' => $VPSID[$vpsid][11],
                'network' => $VPSID[$vpsid][12],
                'hvm' => $VPSID[$vpsid][14],
                'virt_type' => $VPSID[$vpsid][15],
                'autostart' => $VPSID[$vpsid][18],
                'graphics' => $VPSID[$vpsid][19],
                'vnc_listen' => $VPSID[$vpsid][21],
                'vnc_port' => $VPSID[$vpsid][22],
                'status' => $VPSID[$vpsid][25],
                'is_dir_present' => $VPSID[$vpsid][26],
                'is_vps_present' => $VPSID[$vpsid][27],
                'cpuTime' => $VPSID[$vpsid][28],
                'currMem' => $VPSID[$vpsid][29],
                'disk' => $VPSID[$vpsid][31],
                'disk_alloc' => $VPSID[$vpsid][32],
                'timestamp' => time()
                });

		# We may not have a 'vnc_pass' set, so we set it separately if there is one:
        	if ($VPSID[$vpsid][20] ne "") {
        	    ($ok) = $cce->set($oids[0], '',{ 	# beware: CLASS vs. Namespace (see schema)! Our NameSpace is empty, hence: ''
            	    'vnc_pass' => $VPSID[$vpsid][20],
            	    'timestamp' => time()
            	    });
        	}
            }
}

$cce->bye('SUCCESS');
exit(0);

# Legend of Matrix columns:
#
# 0 - vpsid (removed / inactive)
# 1 - vpsname
# 2 - vcpus
# 3 - memory
# 4 - arch
# 5 - machine *
# 6 - iso_source *
# 7 - os_type *
# 8 - os_variant
# 9 - disk_device
#10 - disk_bus
#11 - disk_size
#12 - network
#13 - network_model *
#14 - hvm
#15 - virt_type
#16 - noapic *
#17 - noacpi *
#18 - autostart 
#19 - graphics 
#20 - vnc_pass
#21 - vnc_listen
#22 - vnc_port
#23 - ip *
#24 - hostname *
#25 - status 
#26 - is_dir_present
#27 - is_vps_present
#28 - cpuTime
#29 - currMem
#30 - autostart * (double)
#31 - disk
#32 - disk_alloc
#
# * = We don't get that info by polling the KVM config. Either not at all, or not easily.

